{ *********************************************************************** }
{                                                                         }
{ Adit Math Engine                                                        }
{ Math Demo                                                               }
{                                                                         }
{ Copyright (c) 2015-21 Adit Software http://www.aditmath.com/            }
{                                                                         }
{ *********************************************************************** }


unit frmMain;

interface

uses
  Winapi.Windows, Winapi.Messages, Winapi.ShellApi,
  System.Types, System.SysUtils, System.Variants, System.Classes, System.Math,
  Vcl.Graphics, Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.StdCtrls, Vcl.Grids,
  Vcl.Themes, Vcl.ComCtrls, Vcl.ExtCtrls, Vcl.Clipbrd, Vcl.Menus, Vcl.CheckLst,
  MathImages, MathBase, MathTable, MathParser, MathPresentation, MathEquation,
  MathComponent, MathMacroUtils, MathEdit,  MathScroll;

type
  TMainForm = class(TForm)
    dgSymbols: TDrawGrid;
    memText: TMemo;
    btnUpdate: TButton;
    mc: TMathControl;
    pgcMain: TPageControl;
    tsSymbols: TTabSheet;
    pnlClient: TPanel;
    pnlBottom: TPanel;
    splBottom: TSplitter;
    pnlSymbolsTop: TPanel;
    cbFilter: TComboBox;
    splPc: TSplitter;
    tsExpr: TTabSheet;
    dgExpr: TDrawGrid;
    Panel1: TPanel;
    cbExpr: TComboBox;
    chkLiveUpdate: TCheckBox;
    tsOptions: TTabSheet;
    cbMathFont: TComboBox;
    Label1: TLabel;
    chkInline: TCheckBox;
    cbMathFontSize: TComboBox;
    Label2: TLabel;
    pmMath: TPopupMenu;
    miCopyBitmap: TMenuItem;
    sb: TStatusBar;
    cd: TColorDialog;
    btnFontColor: TButton;
    btnBackColor: TButton;
    miCopyMathJax: TMenuItem;
    miCopyLatex: TMenuItem;
    miCopyMathML: TMenuItem;
    miCopyOMML: TMenuItem;
    miPaste: TMenuItem;
    miCopyMathMLNative: TMenuItem;
    chkMultiline: TCheckBox;
    btnHyperlinkColor: TButton;
    tsBreakOptions: TTabSheet;
    chkEnableBreaks: TCheckBox;
    chkEnableAutoBreaks: TCheckBox;
    lblEqWidth: TLabel;
    cbEqWidth: TComboBox;
    edtEqWidth: TEdit;
    lblBreakAlign: TLabel;
    cbBreakAlign: TComboBox;
    chkBreakBinary: TCheckBox;
    chkBreakRel: TCheckBox;
    chkBreakSpace: TCheckBox;
    chkBreakChars: TCheckBox;
    btnAlignPointColor: TButton;
    edtIndent: TEdit;
    lblBreakStyle: TLabel;
    cbBreakStyle: TComboBox;
    lblSubtractBreak: TLabel;
    cbSubtractBreak: TComboBox;
    edtSubtractChars: TEdit;
    lblAutoWrapType: TLabel;
    cbAutoWrapType: TComboBox;
    edtBreakChars: TEdit;
    lblMultiline: TLabel;
    chkShowNonPrinting: TCheckBox;
    lblShowNonPrinting: TLabel;
    clbNonPrinting: TCheckListBox;
    cbSpaceStyle: TComboBox;
    lblSpaceStyle: TLabel;
    tbZoom: TTrackBar;
    lblZoom: TLabel;
    procedure FormCreate(Sender: TObject);
    procedure btnUpdateClick(Sender: TObject);
    procedure memTextChange(Sender: TObject);
    procedure dgSymbolsDrawCell(Sender: TObject; ACol, ARow: Integer;
      Rect: TRect; State: TGridDrawState);
    procedure FormDestroy(Sender: TObject);
    procedure dgSymbolsDblClick(Sender: TObject);
    procedure pgcMainResize(Sender: TObject);
    procedure cbFilterChange(Sender: TObject);
    procedure dgExprDrawCell(Sender: TObject; ACol, ARow: Integer; Rect: TRect;
      State: TGridDrawState);
    procedure dgExprDblClick(Sender: TObject);
    procedure cbExprChange(Sender: TObject);
    procedure chkLiveUpdateClick(Sender: TObject);
    procedure cbMathFontChange(Sender: TObject);
    procedure chkInlineClick(Sender: TObject);
    procedure cbMathFontSizeChange(Sender: TObject);
    procedure dgSymbolsMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure dgExprMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure miCopyBitmapClick(Sender: TObject);
    procedure btnFontColorClick(Sender: TObject);
    procedure btnBackColorClick(Sender: TObject);
    procedure miCopyMathJaxClick(Sender: TObject);
    procedure miCopyLatexClick(Sender: TObject);
    procedure miCopyMathMLClick(Sender: TObject);
    procedure miCopyOMMLClick(Sender: TObject);
    procedure miPasteClick(Sender: TObject);
    procedure mcStatusLineChanged(Sender: TCustomMathControl;
      AIsStatus: boolean; const AStatusLine: string);
    procedure mcChange(Sender: TObject);
    procedure miCopyMathMLNativeClick(Sender: TObject);
    procedure chkMultilineClick(Sender: TObject);
    procedure btnHyperlinkColorClick(Sender: TObject);
    procedure chkEnableAutoBreaksClick(Sender: TObject);
    procedure chkEnableBreaksClick(Sender: TObject);
    procedure btnAlignPointColorClick(Sender: TObject);
    procedure edtIndentChange(Sender: TObject);
    procedure cbBreakAlignChange(Sender: TObject);
    procedure cbEqWidthChange(Sender: TObject);
    procedure edtEqWidthChange(Sender: TObject);
    procedure cbBreakStyleChange(Sender: TObject);
    procedure cbSubtractBreakChange(Sender: TObject);
    procedure edtSubtractCharsChange(Sender: TObject);
    procedure cbAutoWrapTypeChange(Sender: TObject);
    procedure chkBreakClick(Sender: TObject);
    procedure edtBreakCharsChange(Sender: TObject);
    procedure chkShowNonPrintingClick(Sender: TObject);
    procedure clbNonPrintingClickCheck(Sender: TObject);
    procedure cbSpaceStyleChange(Sender: TObject);
    procedure tbZoomChange(Sender: TObject);
    procedure mcNavigateHyperlink(Sender: TCustomMathControl;
      const AHyperlink: string);
  private
    FLockUpdate, FLockControls: boolean;
    FLastSymbolsGridPos, FLastExprGridPos: TGridCoord;
    FLiveUpdate: boolean;
    FEquation: TMathEquation;
    FActiveSymbolType: TMathSymbolType;
    FActiveExpr: TMathExpressionType;
    FExprs: array of integer;
    procedure FillSymbols;
    procedure FillExpr;
    procedure UpdateText;
  public
  end;

var
  MainForm: TMainForm;

implementation

{$R *.dfm}

procedure TMainForm.UpdateText;
var
  Error: string;
  Index, Position: integer;
begin
  if FLockUpdate then Exit;

  FLockUpdate := True;
  try
    if chkMultiline.Checked then
    begin
      if not mc.TrySetList(memText.Lines, Index, Error, Position) then
        sb.Panels[1].Text := Error + Format(' at position: %d:%d', [Index, Position])
      else
        sb.Panels[1].Text  := '';
    end
    else begin
      if not mc.TrySetText(memText.Lines.Text,  Error, Position) then
        sb.Panels[1].Text := Error + Format(' at position: %d', [Position])
      else
        sb.Panels[1].Text  := '';
    end;
  finally
    FLockUpdate := False;
  end;
end;

procedure TMainForm.btnAlignPointColorClick(Sender: TObject);
begin
  cd.Color := mc.EquationProperties.AlignPointColor;
  if cd.Execute then
    mc.EquationProperties.AlignPointColor := cd.Color;
end;

procedure TMainForm.btnBackColorClick(Sender: TObject);
begin
  cd.Color := mc.EquationProperties.BackgroundColor;
  if cd.Execute then
    mc.EquationProperties.BackgroundColor := cd.Color;
end;

procedure TMainForm.btnFontColorClick(Sender: TObject);
begin
  cd.Color := mc.EquationProperties.TextColor;
  if cd.Execute then
    mc.EquationProperties.TextColor := cd.Color;
end;

procedure TMainForm.btnHyperlinkColorClick(Sender: TObject);
begin
  cd.Color := mc.EquationProperties.HyperlinkColor;
  if cd.Execute then
    mc.EquationProperties.HyperlinkColor := cd.Color;
end;

procedure TMainForm.btnUpdateClick(Sender: TObject);
begin
  UpdateText;
end;

procedure TMainForm.FormCreate(Sender: TObject);
var
  I: TMathSymbolType;
  J: TMathExpressionType;
  NPSymbols: TMathNonPrintingSymbols;
  SL: TStringList;
begin
  UpdateText;
  mc.OnStatusChange := mcStatusLineChanged;
  mc.OnChange := mcChange;
  mc.Color := clWindow;

  FLockUpdate := False;
  FLockControls := False;

  cbFilter.Items.Clear;
  for I := Low(TMathSymbolType) to High(TMathSymbolType) do
    cbFilter.Items.AddObject(GetMathSymbolType(I), TObject(I));
  cbFilter.ItemIndex := 0;

  cbExpr.Items.Clear;
  for J := Low(TMathExpressionType) to High(TMathExpressionType) do
    cbExpr.Items.AddObject(GetMathExpressionType(J), TObject(J));
  cbExpr.ItemIndex := 0;

  FLockControls := True;
  try
    cbMathFont.Items.Clear;
    SL := TMathTable.EnumMathFonts;
    try
      cbMathFont.Items.Assign(SL);
      cbMathFont.ItemIndex := cbMathFont.Items.IndexOf(mc.EquationProperties.Fontname);
    finally
      SL.Free;
    end;

    cbMathFontSize.Items.Clear;
    cbMathFontSize.Items.Add('8');
    cbMathFontSize.Items.Add('9');
    cbMathFontSize.Items.Add('10');
    cbMathFontSize.Items.Add('11');
    cbMathFontSize.Items.Add('12');
    cbMathFontSize.Items.Add('14');
    cbMathFontSize.Items.Add('16');
    cbMathFontSize.Items.Add('18');
    cbMathFontSize.Items.Add('20');
    cbMathFontSize.Items.Add('22');
    cbMathFontSize.Items.Add('24');
    cbMathFontSize.Items.Add('26');
    cbMathFontSize.Items.Add('28');
    cbMathFontSize.Items.Add('36');
    cbMathFontSize.Items.Add('48');
    cbMathFontSize.Items.Add('72');
    cbMathFontSize.Text := FloatToStr(mc.EquationProperties.FontsizeDouble / 2);

    chkInline.Checked := mc.DisplayInline;

    chkShowNonPrinting.Checked := mc.EquationProperties.ShowNonPrintingSymbols;

    lblShowNonPrinting.Enabled := chkShowNonPrinting.Checked;
    clbNonPrinting.Enabled := chkShowNonPrinting.Checked;
    btnAlignPointColor.Enabled := chkShowNonPrinting.Checked;
    lblSpaceStyle.Enabled := chkShowNonPrinting.Checked;
    cbSpaceStyle.Enabled := chkShowNonPrinting.Checked;

    NPSymbols := mc.EquationProperties.VisibleNonPrintingSymbols;
    clbNonPrinting.Checked[0] := (mpsSpace in NPSymbols);
    clbNonPrinting.Checked[1] := (mpsZeroSpace in NPSymbols);
    clbNonPrinting.Checked[2] := (mpsAlignPoint in NPSymbols);
    clbNonPrinting.Checked[3] := (mpsBreak in NPSymbols);
    clbNonPrinting.Checked[4] := (mpsDisplayBreak in NPSymbols);
    clbNonPrinting.Checked[5] := (mpsSoftBreak in NPSymbols);
    cbSpaceStyle.ItemIndex := ord(mc.EquationProperties.SpaceDrawStyle);

    cbEqWidth.ItemIndex := ord(mc.DocumentWidthType);
    if mc.DocumentWidthType = dwtCustomSize then
    begin
      cbEqWidth.ItemIndex := 2;
      edtEqWidth.Enabled := True;
      edtEqWidth.Text := mc.DocumentWidth.ToLatex;
    end;

    chkEnableBreaks.Checked := mc.EquationProperties.EnableBreaks;
    chkEnableAutoBreaks.Checked := mc.EquationProperties.EnableAutoBreaks;

    case mc.EquationProperties.BreakAlign of
      aaLeft: cbBreakAlign.ItemIndex := 0;
      aaCenter: cbBreakAlign.ItemIndex := 2;
      aaRight: cbBreakAlign.ItemIndex := 1;
    else
      cbBreakAlign.ItemIndex := -1;
    end;

    case mc.EquationProperties.BreakStyle of
      lbsBefore: cbBreakStyle.ItemIndex := 0;
      lbsAfter: cbBreakStyle.ItemIndex := 1;
      lbsDuplicate: cbBreakStyle.ItemIndex := 2;
    else
      cbBreakStyle.ItemIndex := -1;
    end;

    lblSubtractBreak.Enabled := (mc.EquationProperties.BreakStyle = lbsDuplicate);
    cbSubtractBreak.Enabled := (mc.EquationProperties.BreakStyle = lbsDuplicate);
    edtSubtractChars.Enabled := (mc.EquationProperties.BreakStyle = lbsDuplicate);

    cbSubtractBreak.ItemIndex := ord(mc.EquationProperties.SubtractBreakStyle);
    edtSubtractChars.Text := CharArrToString(mc.EquationProperties.SubtractChars);

    cbAutoWrapType.ItemIndex := ord(mc.EquationProperties.AutoWrapType);
    edtIndent.Enabled := (mc.EquationProperties.AutoWrapType = awtFixedIndent);
    edtIndent.Text := mc.EquationProperties.Indent.ToLatex;

    chkBreakBinary.Checked := absBreakOnBin in mc.EquationProperties.AutoBreakStyles;
    chkBreakRel.Checked := absBreakOnRel in mc.EquationProperties.AutoBreakStyles;
    chkBreakSpace.Checked := absBreakOnSpace in mc.EquationProperties.AutoBreakStyles;
    chkBreakChars.Checked := absBreakOnChar in mc.EquationProperties.AutoBreakStyles;
    edtBreakChars.Enabled := absBreakOnChar in mc.EquationProperties.AutoBreakStyles;
    edtBreakChars.Text:= CharArrToString(mc.EquationProperties.BreakChars);
  finally
    FLockControls := False;
  end;

  FEquation := TMathEquation.Create(nil);
  FEquation.Properties.FontsizeDouble := 24;
  FEquation.DisplayInline := true;
  FActiveSymbolType := mstBasic;
  dgSymbols.DefaultColWidth := GetSymbolSize(GetDeviceCaps(Canvas.Handle, LOGPIXELSY));
  dgSymbols.DefaultRowHeight := dgSymbols.DefaultColWidth;

  FActiveExpr := metFraction;
  SetLength(FExprs, 0);

  FillSymbols;
  FillExpr;
  FLiveUpdate := True;
  UpdateText;
end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
  FreeAndNil(FEquation);
end;

procedure TMainForm.mcStatusLineChanged(Sender: TCustomMathControl;
  AIsStatus: boolean; const AStatusLine: string);
begin
  if AIsStatus then
    sb.Panels[0].Text := AStatusLine
  else
    sb.Panels[0].Text := '';
end;

procedure TMainForm.mcChange(Sender: TObject);
begin
  if FLockUpdate then Exit;

  FLockUpdate := True;
  try
    memText.Text := mc.Text;
    chkInline.Checked := mc.DisplayInline;
  finally
    FLockUpdate := False;
  end;
end;

procedure TMainForm.mcNavigateHyperlink(Sender: TCustomMathControl;
  const AHyperlink: string);
begin
  if ShellExecute(0, 'open', PChar(AHyperlink), nil, nil, SW_SHOW) <= 32 then
    MessageBox(0, PChar(Format('Can not navigate hyperlink: %s', [AHyperlink])), PChar('Error'), MB_OK or MB_ICONHAND);
end;

procedure TMainForm.cbAutoWrapTypeChange(Sender: TObject);
begin
  if FLockControls then Exit;
  mc.EquationProperties.AutoWrapType := TAutoWrapType(cbAutoWrapType.ItemIndex);
  edtIndent.Enabled := (cbAutoWrapType.ItemIndex = 1);
end;

procedure TMainForm.cbBreakAlignChange(Sender: TObject);
var
  Align: TAlignment;
begin
  if FLockControls then Exit;
  if cbBreakAlign.ItemIndex >= 0 then
  begin
    Align := TAlignment(cbBreakAlign.ItemIndex);
    case Align of
      taLeftJustify: mc.EquationProperties.BreakAlign := aaLeft;
      taRightJustify: mc.EquationProperties.BreakAlign := aaRight;
      taCenter: mc.EquationProperties.BreakAlign := aaCenter;
    end;
  end;
end;

procedure TMainForm.cbBreakStyleChange(Sender: TObject);
begin
  if FLockControls then Exit;

  case cbBreakStyle.ItemIndex of
    0: mc.EquationProperties.BreakStyle := lbsBefore;
    1: mc.EquationProperties.BreakStyle := lbsAfter;
    2: mc.EquationProperties.BreakStyle := lbsDuplicate;
  end;

  lblSubtractBreak.Enabled := (cbBreakStyle.ItemIndex = 2);
  cbSubtractBreak.Enabled := (cbBreakStyle.ItemIndex = 2);
  edtSubtractChars.Enabled := (cbBreakStyle.ItemIndex = 2);
end;

procedure TMainForm.cbEqWidthChange(Sender: TObject);
var
  Sz: TTeXSize;
begin
  if FLockControls then Exit;
  mc.DocumentWidthType :=  TDocumentWidthType(cbEqWidth.ItemIndex);
  edtEqWidth.Enabled := (cbEqWidth.ItemIndex > 2);
  if (cbEqWidth.ItemIndex > 2) and TryStrToTeXSize(edtEqWidth.Text, False, True, False, Sz) then
    mc.DocumentWidth := Sz;
end;

procedure TMainForm.cbExprChange(Sender: TObject);
begin
  if cbExpr.ItemIndex >= 0 then
  begin
    FActiveExpr := TMathExpressionType(cbExpr.Items.Objects[cbExpr.ItemIndex]);
    FillExpr;
    dgExpr.Invalidate;
  end;
end;

procedure TMainForm.cbFilterChange(Sender: TObject);
begin
  if cbFilter.ItemIndex >= 0 then
  begin
    FActiveSymbolType := TMathSymbolType(cbFilter.Items.Objects[cbFilter.ItemIndex]);
    FillSymbols;
    dgSymbols.Invalidate;
  end;
end;

procedure TMainForm.cbMathFontChange(Sender: TObject);
var
  Fontname: string;
begin
  if FLockControls then Exit;
  if cbMathFont.ItemIndex >= 0 then
  begin
    Fontname := cbMathFont.Items[cbMathFont.ItemIndex];
    if TMathTable.IsValidMathFont(Fontname) then
      mc.EquationProperties.Fontname := Fontname;
  end;
end;

procedure TMainForm.cbMathFontSizeChange(Sender: TObject);
var
  Sz: double;
begin
  if FLockControls then Exit;
  if TryStrToFloat(cbMathFontSize.Text, Sz) and (Sz > 0) and (Sz < 500) then
    mc.EquationProperties.FontsizeDouble := Round(Sz * 2);
end;

procedure TMainForm.cbSpaceStyleChange(Sender: TObject);
begin
  if FLockControls then Exit;
  mc.EquationProperties.SpaceDrawStyle := TMathSpaceDrawStyle(cbSpaceStyle.ItemIndex);
end;

procedure TMainForm.cbSubtractBreakChange(Sender: TObject);
begin
  if FLockControls then Exit;
  mc.EquationProperties.SubtractBreakStyle := TSubtractBreakStyle(cbSubtractBreak.ItemIndex);
end;

procedure TMainForm.chkBreakClick(Sender: TObject);
var
  Styles: TAutoBreakStyles;
begin
  if FLockControls then Exit;
  if chkBreakBinary.Checked then Include(Styles, absBreakOnBin);
  if chkBreakRel.Checked then Include(Styles, absBreakOnRel);
  if chkBreakSpace.Checked then Include(Styles, absBreakOnSpace);
  if chkBreakChars.Checked then Include(Styles, absBreakOnChar);
  mc.EquationProperties.AutoBreakStyles := Styles;
  edtBreakChars.Enabled := absBreakOnChar in Styles;
end;

procedure TMainForm.chkEnableAutoBreaksClick(Sender: TObject);
begin
  if FLockControls then Exit;
  mc.EquationProperties.EnableAutoBreaks := chkEnableAutoBreaks.Checked;
end;

procedure TMainForm.chkEnableBreaksClick(Sender: TObject);
begin
  if FLockControls then Exit;
  mc.EquationProperties.EnableBreaks := chkEnableBreaks.Checked;
end;

procedure TMainForm.chkInlineClick(Sender: TObject);
begin
  if FLockControls then Exit;
  mc.DisplayInline := chkInline.Checked;
end;

procedure TMainForm.edtBreakCharsChange(Sender: TObject);
begin
  if FLockControls then Exit;
  mc.EquationProperties.BreakChars := StringToCharArr(edtBreakChars.Text);
end;

procedure TMainForm.edtIndentChange(Sender: TObject);
var
  Sz: TTeXSize;
begin
  if FLockControls then Exit;
  if TryStrToTeXSize(edtIndent.Text, False, True, False, Sz) then
    mc.EquationProperties.Indent := Sz;
end;

procedure TMainForm.edtSubtractCharsChange(Sender: TObject);
begin
  if FLockControls then Exit;
  mc.EquationProperties.SubtractChars := StringToCharArr(edtSubtractChars.Text);
end;

procedure TMainForm.edtEqWidthChange(Sender: TObject);
var
  Sz: TTeXSize;
begin
  if FLockControls then Exit;
  if TryStrToTeXSize(edtEqWidth.Text, False, True, False, Sz) then
    mc.DocumentWidth := Sz;
end;

procedure TMainForm.tbZoomChange(Sender: TObject);
begin
  mc.Zoom := tbZoom.Position / 100;
end;

procedure TMainForm.chkLiveUpdateClick(Sender: TObject);
begin
  FLiveUpdate := chkLiveUpdate.Checked;
end;

procedure TMainForm.chkMultilineClick(Sender: TObject);
begin
  mc.ForceContentType(chkMultiline.Checked);
  memText.WordWrap := not chkMultiline.Checked;
end;

procedure TMainForm.chkShowNonPrintingClick(Sender: TObject);
begin
  if FLockControls then Exit;
  lblShowNonPrinting.Enabled := chkShowNonPrinting.Checked;
  clbNonPrinting.Enabled := chkShowNonPrinting.Checked;
  btnAlignPointColor.Enabled := chkShowNonPrinting.Checked;
  lblSpaceStyle.Enabled := chkShowNonPrinting.Checked;
  cbSpaceStyle.Enabled := chkShowNonPrinting.Checked;

  mc.EquationProperties.ShowNonPrintingSymbols := chkShowNonPrinting.Checked;
end;

procedure TMainForm.clbNonPrintingClickCheck(Sender: TObject);
var
  NPSymbols: TMathNonPrintingSymbols;
begin
  if FLockControls then Exit;

  NPSymbols := [];
  if clbNonPrinting.Checked[0] then Include(NPSymbols, mpsSpace);
  if clbNonPrinting.Checked[1] then Include(NPSymbols, mpsZeroSpace);
  if clbNonPrinting.Checked[2] then Include(NPSymbols, mpsAlignPoint);
  if clbNonPrinting.Checked[3] then Include(NPSymbols, mpsBreak);
  if clbNonPrinting.Checked[4] then Include(NPSymbols, mpsDisplayBreak);
  if clbNonPrinting.Checked[5] then Include(NPSymbols, mpsSoftBreak);
  mc.EquationProperties.VisibleNonPrintingSymbols := NPSymbols;
end;

procedure TMainForm.dgExprDblClick(Sender: TObject);
var
  Index, I: integer;
  Command: string;
begin
  Index := dgExpr.Row * dgExpr.ColCount + dgExpr.Col;
  if (Index >= 0) and (Index <= High(FExprs)) then
  begin
    Command := MathExpressions[FExprs[Index]].Command;
    if HavePlaceholder(Command) then
    begin
      if memText.SelText <> '' then
      begin
        Command := StringReplace(Command, '%%', memText.SelText, [rfReplaceAll]);
        I := Pos('%', Command);
        Command := StringReplace(Command, '%', '', [rfReplaceAll]);
      end
      else begin
        I := Pos('%%', Command);
        if I > 1 then
        begin
          Command := StringReplace(Copy(Command, 1, I-1), '%', '', [rfReplaceAll]) +
            Copy(Command, I, MaxInt);
          I := Pos('%%', Command);
        end;

        Command := StringReplace(Command, '%%', '', [rfReplaceAll]);
        Command := StringReplace(Command, '%', '', [rfReplaceAll]);
      end;

      memText.SelText := Command;
      if I > 0 then
        memText.SelStart := memText.SelStart + I - 1;
      memText.SelLength := 0;
    end
    else
      memText.SelText := Command;
  end;
  memText.SetFocus;
end;

procedure TMainForm.dgExprDrawCell(Sender: TObject; ACol, ARow: Integer;
  Rect: TRect; State: TGridDrawState);
var
  Details: TThemedGrid;
  Index, W, H: integer;
  Err, Command: string;
  Pos: integer;
  Color: TColor;
  ActiveRange: TMathSymbolRanges;
begin
  Index := ARow * dgExpr.ColCount + ACol;

  if gdSelected in State then
    Details := tgCellSelected
  else
    Details := tgCellNormal;
  if not (gdSelected in State) or not (ThemeControl(Self) and
    StyleServices.DrawElement(dgExpr.Canvas.Handle,
      StyleServices.GetElementDetails(Details), Rect)) then
  begin
    if gdSelected in State then
      dgExpr.Canvas.Brush.Color := clHighlight
    else
      if StyleServices.Enabled then
        dgExpr.Canvas.Brush.Color := StyleServices.GetStyleColor(scGrid)
      else
        dgExpr.Canvas.Brush.Color := clWindow;
    dgExpr.Canvas.FillRect(Rect);
  end;

  if ThemeControl(Self) and StyleServices.GetElementColor(StyleServices.GetElementDetails(Details), ecTextColor, Color) and
    (Color <> clNone) then
    FEquation.Properties.TextColor := Color
  else
  begin
    if gdSelected in State then
      FEquation.Properties.TextColor := clHighlightText
    else
      FEquation.Properties.TextColor := clWindowText;
  end;

  ActiveRange := MathRanges[FActiveSymbolType];
  if (Index >= 0) and (Index <= High(FExprs)) then
  begin
    Command := MathExpressions[FExprs[Index]].Command;
    if HavePlaceholder(Command) then
    begin
      Command := StringReplace(Command, '%%', '\dottedbox', [rfReplaceAll]);
      Command := StringReplace(Command, '%', '\dottedbox', [rfReplaceAll]);
    end;

    FEquation.FromLatexSingleline(Command, Err, Pos);
    FEquation.PrepareDraw(dgExpr.Canvas.Handle);
    W := FEquation.Layout.Width;
    H := FEquation.Layout.Height + FEquation.Layout.Depth;
    FEquation.Draw(dgExpr.Canvas.Handle,
      Rect.Left + (Rect.Right - Rect.Left - W) div 2,
      Rect.Top + (Rect.Bottom - Rect.Top - H) div 2 + FEquation.Layout.Height);
  end;
end;

procedure TMainForm.dgExprMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
var
  GridPos: TGridCoord;
  Index: integer;
begin
  GridPos := dgExpr.MouseCoord(X, Y);
  Index := GridPos.Y * dgExpr.ColCount + GridPos.X;
  if (GridPos.Y <> FLastExprGridPos.Y) or (GridPos.X <> FLastExprGridPos.X) then
  begin
    Application.CancelHint;

  if (Index >= 0) and (Index <= High(FExprs)) then
    begin
      dgExpr.Hint := MathExpressions[FExprs[Index]].Hint;
      FLastExprGridPos := GridPos;
    end;
  end;
end;

procedure TMainForm.dgSymbolsDblClick(Sender: TObject);
var
  Index: integer;
  ActiveRange: TMathSymbolRanges;
begin
  Index := dgSymbols.Row * dgSymbols.ColCount + dgSymbols.Col;
  ActiveRange := MathRanges[FActiveSymbolType];
  if (Index >= 0) and (Index <= High(ActiveRange.ImageIndexes)) and (ActiveRange.ImageIndexes[Index] <> -1)  then
  begin
    memText.SelText := MathSymbols[ActiveRange.ImageIndexes[Index] - 1].Command;
    memText.SetFocus;
  end;
end;

procedure TMainForm.dgSymbolsDrawCell(Sender: TObject; ACol, ARow: Integer;
  Rect: TRect; State: TGridDrawState);
var
  Details: TThemedGrid;
  Index, W, H: integer;
  Err: string;
  Pos: integer;
  Color: TColor;
  ActiveRange: TMathSymbolRanges;
begin
  Index := ARow * dgSymbols.ColCount + ACol;

  if gdSelected in State then
    Details := tgCellSelected
  else
    Details := tgCellNormal;
  if not (gdSelected in State) or not (ThemeControl(Self) and
    StyleServices.DrawElement(dgSymbols.Canvas.Handle,
      StyleServices.GetElementDetails(Details), Rect)) then
  begin
    if gdSelected in State then
      dgSymbols.Canvas.Brush.Color := clHighlight
    else
      if StyleServices.Enabled then
        dgSymbols.Canvas.Brush.Color := StyleServices.GetStyleColor(scGrid)
      else
        dgSymbols.Canvas.Brush.Color := clWindow;
    dgSymbols.Canvas.FillRect(Rect);
  end;

  if ThemeControl(Self) and StyleServices.GetElementColor(StyleServices.GetElementDetails(Details), ecTextColor, Color) and
    (Color <> clNone) then
    FEquation.Properties.TextColor := Color
  else
  begin
    if gdSelected in State then
      FEquation.Properties.TextColor := clHighlightText
    else
      FEquation.Properties.TextColor := clWindowText;
  end;

  ActiveRange := MathRanges[FActiveSymbolType];
  if (Index >= 0) and (Index <= High(ActiveRange.ImageIndexes)) and (ActiveRange.ImageIndexes[Index] <> -1)  then
  begin
    FEquation.FromLatexSingleline(MathSymbols[ActiveRange.ImageIndexes[Index] - 1].Command, Err, Pos);
    FEquation.PrepareDraw(dgSymbols.Canvas.Handle);
    W := FEquation.Layout.Width;
    H := FEquation.Layout.Height + FEquation.Layout.Depth;
    FEquation.Draw(dgSymbols.Canvas.Handle,
      Rect.Left + (Rect.Right - Rect.Left - W) div 2,
      Rect.Top + (Rect.Bottom - Rect.Top - H) div 2 + FEquation.Layout.Height);
  end;
end;

procedure TMainForm.dgSymbolsMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
var
  GridPos: TGridCoord;
  Index: integer;
  ActiveRange: TMathSymbolRanges;
begin
  GridPos := dgSymbols.MouseCoord(X, Y);
  Index := GridPos.Y * dgSymbols.ColCount + GridPos.X;
  if (GridPos.Y <> FLastSymbolsGridPos.Y) or (GridPos.X <> FLastSymbolsGridPos.X) then
  begin
    Application.CancelHint;

    ActiveRange := MathRanges[FActiveSymbolType];
    if (Index >= 0) and (Index <= High(ActiveRange.ImageIndexes)) and (ActiveRange.ImageIndexes[Index] <> -1)  then
    begin
      dgSymbols.Hint := MathSymbols[ActiveRange.ImageIndexes[Index] - 1].Hint;
      FLastSymbolsGridPos := GridPos;
    end;
  end;
end;

procedure TMainForm.FillSymbols;
var
  ActiveRange: TMathSymbolRanges;
  I, SymbolsCount: integer;
begin
  FLastSymbolsGridPos.X := -1;
  FLastSymbolsGridPos.Y := -1;

  ActiveRange := MathRanges[FActiveSymbolType];
  SymbolsCount := 0;
  for I := High(ActiveRange.ImageIndexes) downto 0 do
    if ActiveRange.ImageIndexes[I] <> -1 then
    begin
      SymbolsCount := I+1;
      Break;
    end;

  dgSymbols.ColCount := Max(dgSymbols.ClientWidth div (dgSymbols.DefaultColWidth + 2 * dgSymbols.GridLineWidth ), 1);
  dgSymbols.RowCount := Ceil(SymbolsCount / dgSymbols.ColCount);
end;

procedure TMainForm.FillExpr;
var
  I, ExprCount: integer;
  ppi: integer;
  Size, MaxSize: TPoint;
begin
  FLastExprGridPos.X := -1;
  FLastExprGridPos.Y := -1;
  ExprCount := 0;
  MaxSize := Point(0, 0);
  SetLength(FExprs, 50);
  ppi := GetDeviceCaps(Canvas.Handle, LOGPIXELSY);
  for I := Low(MathExpressions) to High(MathExpressions) do
    if MathExpressions[I].ExprType = FActiveExpr then
    begin
      FExprs[ExprCount] := I;
      Inc(ExprCount);
      Size := GetExpressionSize(ppi, MathExpressions[I].Size);
      if Size.X > MaxSize.X then MaxSize.X := Size.X;
      if Size.Y > MaxSize.Y then MaxSize.Y := Size.Y;
    end;

  SetLength(FExprs, ExprCount);
  dgExpr.DefaultColWidth := Size.X;
  dgExpr.DefaultRowHeight := Size.Y;

  dgExpr.ColCount := Max(dgExpr.ClientWidth div (dgExpr.DefaultColWidth + 2 * dgExpr.GridLineWidth ), 1);
  dgExpr.RowCount := Ceil(ExprCount / dgExpr.ColCount);
end;

procedure TMainForm.memTextChange(Sender: TObject);
begin
  if FLiveUpdate then UpdateText;
end;

procedure TMainForm.miCopyBitmapClick(Sender: TObject);
var
  bmp: TBitmap;
begin
  bmp := mc.AsBitmap;
  try
    Clipboard.Assign(bmp);
  finally
    bmp.Free;
  end;
end;

procedure TMainForm.miCopyLatexClick(Sender: TObject);
begin
  mc.TextCopyFormat := mtcfLatexClassic;
  mc.Copy;
end;

procedure TMainForm.miCopyMathJaxClick(Sender: TObject);
begin
  mc.TextCopyFormat := mtcfLatexMathJax;
  mc.Copy;
end;

procedure TMainForm.miCopyMathMLClick(Sender: TObject);
begin
  mc.TextCopyFormat := mtcfMathMLOffice;
  mc.Copy;
end;

procedure TMainForm.miCopyMathMLNativeClick(Sender: TObject);
begin
  mc.TextCopyFormat := mtcfMathMLClassic;
  mc.Copy;
end;

procedure TMainForm.miCopyOMMLClick(Sender: TObject);
begin
  Clipboard.AsText := mc.AsOMML(True, True);
end;

procedure TMainForm.miPasteClick(Sender: TObject);
begin
  mc.Paste;
end;

procedure TMainForm.pgcMainResize(Sender: TObject);
begin
  FillSymbols;
  FillExpr;
end;

end.
